-- SHIFTJIS.TXT file was obtained from:
-- https://unicode.org/Public/MAPPINGS/OBSOLETE/EASTASIA/JIS/SHIFTJIS.TXT
local utf8 = require'utf8'
local jis2uni = {}
local mcdata, canvas
local deffnt, cjkfnt
local lh = 38 -- Line height
local lo = 30 -- Line offset
local hy = 6 -- Header Y position
local frame2 = 1
local frame3 = 1
local lasttime2
local lasttime3
local icons = {}
local lookup = {}

do
  local f = love.filesystem.read('SHIFTJIS.TXT')

  for i in f:gmatch("\n([^#]+)") do
    local jis, uni = i:match('^([^%s]+)%s+([^%s]+)')
    local byte1 = tonumber(jis:sub(1, 4))
    if #jis == 6 then
      local byte2 = tonumber('0x' .. jis:sub(5, 6))
      if not jis2uni[byte1] then
        jis2uni[byte1] = {}
      end
      jis2uni[byte1][byte2] = tonumber(uni)
    else
      jis2uni[byte1] = tonumber(uni)
    end
  end

  f = love.filesystem.read('serials.txt')

  for i in f:gmatch("[^\n]+") do
    local serial, name = i:match('^([^ ]+) (.*)$')
    lookup[serial] = name
  end
end

local function sjis2utf8(str)
  local i = 1
  local strlen = #str
  local ret = ''
  while i <= strlen do
    local byte1 = str:byte(i)
    if byte1 == 0 then
      break
    end
    if jis2uni[byte1] then
      if type(jis2uni[byte1]) == 'table' then
        i = i + 1
        local byte2 = str:byte(i)
        if not byte2 then
          return ret
        end
        ret = ret .. utf8.char(jis2uni[byte1][byte2])
      else
        ret = ret .. utf8.char(jis2uni[byte1])
      end
    else
      ret = ret .. utf8.char(byte1)
    end
    i = i + 1
  end
  return ret
end

local function latin1_to_utf8(str)
  local needsconv = false
  local strlen = #str
  for i = 1, strlen do
    if str:byte(i) > 128 then
      needsconv = true
      break
    end
  end
  if not needsconv then
    -- most common case
    return str
  end
  for i = 1, strlen do
    ret = ret .. utf8.char(str:byte(i))
  end
  return ret
end

--love.event.quit()

function love.conf(c)
  c.window = nil
end

function love.keypressed(k)
  return k == "escape" and love.event.quit()
end

local function badformat()
  error("Memory card format is wrong or not understood")
end


local function renderTitle(i)
  local blocktype = mcdata:byte(i*128+1)
  local rcode = mcdata:sub(i*128+11, i*128+12)
  local iserial = mcdata:sub(i*128+13, i*128+22)
  local saveid = mcdata:sub(i*128+23, i*128+30)
  local title = mcdata:sub(i*8192+5, i*8192+68)
  local nicons = mcdata:byte(i*8192+3)
  local nblocks = mcdata:byte(i*8192+4)
  local linkno = mcdata:byte(i*128+9) + 256*mcdata:byte(i*128+10)
  if linkno > 32767 then
    linkno = linkno - 65536
  end

  -- Sanity checks
  if not iserial:match('^[A-Za-z0-9%-]+$') then
    badformat()
  end
  if not (nicons == 0 or nicons >= 0x11 and nicons <= 0x13) then
    badformat()
  end
  -- Check size in bytes of save against # blocks used
  if mcdata:byte(i*128+5) + 256*mcdata:byte(i*128+6) + 65536*mcdata:byte(i*128+7) ~= nblocks*8192 then
print(mcdata:byte(i*128+5) + 256*mcdata:byte(i*128+6) + 65536*mcdata:byte(i*128+7), nblocks)
    badformat()
  end

  game = lookup[iserial:sub(1, 4) .. '-' .. iserial:sub(6)] or '???'

  if rcode == 'BI' then
    rcode = 'Japan'
  elseif rcode == 'BE' then
    rcode = 'Europe'
  elseif rcode == 'BA' then
    rcode = 'America'
  else
    rcode = ('Region 0x%02X%02X'):format(rcode:byte(1), rcode:byte(2))
  end

  nicons = nicons % 16
  icons[i] = {}
  for j = 1, nicons do
    icons[i][j] = love.image.newImageData(16, 16)
    icons[i][j]:mapPixel(function(x, y)
      local oddx = x % 2
      local pix = mcdata:byte(i*8192+j*128+y*8+(x-oddx)*0.5+1)
      local plo = pix % 16
      if oddx == 1 then pix = (pix - plo) / 16 else pix = plo end
      local pal1 = mcdata:byte(i*8192 + pix*2 + 97)
      local pal2 = mcdata:byte(i*8192 + pix*2 + 98)
      local palR = pal1 % 32
      local palG = pal2 % 4
      local palB = ((pal2 - palG) * 0.25) % 32
      palG = palG * 8 + (pal1 - palR) / 32
      palR = (palR/31) ^ 1
      palG = (palG/31) ^ 1
      palB = (palB/31) ^ 1
      return palR, palG, palB, pal1+pal2 == 0 and 0 or 1
    end)
    -- TODO: 16 pixels x 3 icons x 15 saves = 720 pixels. An image
    -- of 16x720 or 720x16 could be used as texture atlas instead.
    icons[i][j] = love.graphics.newImage(icons[i][j])
    icons[i][j]:setFilter("nearest", "nearest")
  end

print("Block " .. i .. ", " .. nblocks .. " blocks occupied, links to " .. linkno+1)
  love.graphics.print("Block " .. i, 10, (i-1)*lh+lo)
  love.graphics.print(tostring(nblocks), 70, (i-1)*lh+lo)
  love.graphics.print(tostring(linkno+1), 95, (i-1)*lh+lo)
print(latin1_to_utf8(iserial) .. ' - ' .. game)
  love.graphics.print(latin1_to_utf8(iserial), 120, (i-1)*lh+lo)
print("Region: " .. rcode)
  love.graphics.print(rcode, 200, (i-1)*lh+lo)
  saveid = saveid:match("^(.-)%z*$")
print("SaveID: " .. latin1_to_utf8(saveid))
  love.graphics.print(latin1_to_utf8(saveid), 260, (i-1)*lh+lo)
print("Title: " .. sjis2utf8(title))
print()
  love.graphics.setFont(cjkfnt)
  love.graphics.print(sjis2utf8(title), 340, (i-1)*lh+lo)
  love.graphics.setFont(deffnt)
  love.graphics.print(game, 10, (i-1)*lh + lo + 16)
  if nicons == 1 then
    love.graphics.draw(icons[i][1], 758, 3 + (i-1)*lh + lo, 0, 2, 2)
  end
end


local function renderLink(i)
  local blocktype = mcdata:byte(i*128+1)
  local linkno = mcdata:byte(i*128+9) + 256*mcdata:byte(i*128+10)
  if linkno == 65535 then
print("Block " .. i .. ": EOF")
    love.graphics.print("Block " .. i .. " is end of chain", 10, (i-1)*lh+lo)
    if blocktype ~= 0x53 then
      badformat()
    end
  else
print("Block " .. i .. " links to " .. linkno+1)
    love.graphics.print("Block " .. i .. " links to block " .. linkno+1,  10, (i-1)*lh+lo)
    if blocktype ~= 0x52 then
      badformat()
    end
  end
end

local function renderEmpty(i)
  love.graphics.print("Block " .. i .. " empty", 10, (i-1)*lh+lo)
end


local function renderBad(i)
print("Bad block: " .. i)
  love.graphics.print("Block " .. i .. " is unusable", 10, (i-1)*lh+lo)
end


local function renderHeading(bg, q)
  love.graphics.draw(bg, q, 0, 0, 0, 800, lo)
  love.graphics.print("BlockNo", 10, hy)
  love.graphics.print("#B", 70, hy)
  love.graphics.print("Lk", 95, hy)
  love.graphics.print("Serial", 120, hy)
  love.graphics.print("Region", 200, hy)
  love.graphics.print("Save ID", 260, hy)
  love.graphics.print("Title", 340, hy)
  love.graphics.print("Icon", 758, hy)
end


function love.load(args)
  do
    if not args[1] then
      io.stdout:write("Usage: love mclist.love <memcard_path>\n")
      return love.event.quit()
    end
    local f = io.open(args[1], "rb")
    if f == nil then
      io.stderr:write("Error reading ", args[1], "\n")
      return love.event.quit()
    end
    local ok
    ok, mcdata = pcall(f.read, f, "*all")
    f:close()
    assert(ok, mcdata)

    love.window.setMode(800, 600, {resizable=true})
    lasttime2 = love.timer.getTime()
    lasttime3 = lasttime2
    deffnt = love.graphics.getFont()
--    cjkfnt = love.graphics.newFont('NotoSansJP[wght].ttf')
--    cjkfnt = love.graphics.newFont('NotoSansJP[wght].ttf')
--    cjkfnt = love.graphics.newFont('NotoSansJP-Medium.otf')
    cjkfnt = love.graphics.newFont('NotoSansCJKjp-Regular.otf')
    deffnt = cjkfnt
    canvas = love.graphics.newCanvas()
    love.graphics.setCanvas(canvas)
    love.graphics.setFont(deffnt)

    local bgimage = love.graphics.newImage(love.image.newImageData(
      1, 2, "rgba8", "\x30\x50\x70\xFF\x20\x40\x60\xFF"))
    local bghimage = love.graphics.newImage(love.image.newImageData(
      1, 2, "rgba8", "\x50\x50\x50\xFF\x40\x40\x40\xFF"))
    local bgquad = love.graphics.newQuad(0, 0.5, 1, 1, 1, 2)

    renderHeading(bghimage, bgquad)

    for i = 1, 15 do
      love.graphics.draw(bgimage, bgquad, 0, (i-1)*lh + lo, 0, 800, lh)
      icons[i] = false
      local blocktype = mcdata:byte(i*128+1)
      if blocktype == 0x51 then
        if mcdata:byte(i*8192+1) ~= 0x53 or mcdata:byte(i*8192+2) ~= 0x43 then
          badformat()
        end
        renderTitle(i)
      elseif blocktype == 0x52 or blocktype == 0x53 then
        renderLink(i)
      elseif blocktype >= 0xF0 then
        renderBad(i)
      elseif blocktype >= 0xA0 and blocktype <= 0xAF then
        renderEmpty(i)
      else
        badformat()
      end
    end
    love.graphics.setCanvas()
  end
  collectgarbage('collect')
  love.window.setTitle(args[1] .. " - mclist")
end

function love.draw()
  -- Render static stuff
  love.graphics.setBlendMode("alpha", "premultiplied")
  love.graphics.draw(canvas)
  love.graphics.setBlendMode("alpha", "alphamultiply")

  local newtime = love.timer.getTime()
  if newtime - lasttime2 >= 0.32 then
    frame2 = frame2 % 2 + 1
    lasttime2 = lasttime2 + 0.32
  end
  if newtime - lasttime3 >= 0.22 then
    frame3 = frame3 % 3 + 1
    lasttime3 = lasttime3 + 0.22
  end

  -- Render icons
  for i = 1, 15 do
    local icon = false
    local iconframes = icons[i]
    if iconframes then
      local nicons = #iconframes
      if nicons == 2 then
        icon = iconframes[frame2]
      elseif nicons == 3 then
        icon = iconframes[frame3]
      end
      if icon then
        love.graphics.draw(icon, 758, 3 + (i-1)*lh + lo, 0, 2, 2)
      end
    end
  end
end
